# -*- coding: utf-8 -*-
"""
This package contains the Universal Ink Model codec and a decoder for the Wacom Ink Layer Language (WILL) format.
"""
__all__ = ['context', 'parser', 'writer']

from uim.codec import context
from uim.codec import parser
from uim.codec import writer
