# -*- coding: utf-8 -*-
"""
Semantics
=========
The semantics package contains the following modules:
- node: The node module contains the base classes for the Universal Ink Model (UIM) nodes.
- structures: The structures module contains the data structures used within the Universal Ink Model (UIM).
- schema: The schema module contains the schema classes for the Universal Ink Model (UIM).
"""

__all__ = ['node', 'structures', 'schema']
